import argparse
import pathlib
import sys

import lmfit as lf
import nmrglue as ng
import numpy as np

from peakfit import clustering, computing, monte_carlo, shapes, __version__
from peakfit.util import print_peaks


LOGO = """

* * * * * * * * * * * * * * * * * * * * * * 
*    ____            _    _____ _ _       *
*   |  _ \ ___  __ _| | _|  ___(_) |_     *
*   | |_) / _ \/ _` | |/ / |_  | | __|    *
*   |  __/  __/ (_| |   <|  _| | | |_     *
*   |_|   \___|\__,_|_|\_\_|   |_|\__|    *
*                                         *
*  Perform peak integration in pseudo-3D  *
*  spectra                                *
*                                         *
*  Version: {:<29s} *
*                                         *
* * * * * * * * * * * * * * * * * * * * * *
""".format(
    __version__
)


def parse_command_line():
    description = "Perform peak integration in pseudo-3D."

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("-s", dest="path_spectra", required=True)
    parser.add_argument("-l", dest="path_list_peak", required=True)
    parser.add_argument("-z", dest="path_list_z", required=True)
    parser.add_argument("-t", dest="contour_level", required=True, type=float)
    parser.add_argument("-o", dest="path_output", default="Fits", type=pathlib.Path)
    parser.add_argument("-n", dest="noise", default=1.0, type=float)
    parser.add_argument(
        "--mc",
        dest="noise_box",
        default=None,
        nargs=5,
        metavar=("X_MIN", "X_MAX", "Y_MIN", "Y_MAX", "N"),
    )

    return parser.parse_args()


def main():

    print(LOGO)

    args = parse_command_line()

    # Read spectra
    dic, spectra = ng.fileio.pipe.read(args.path_spectra)

    # Read peak list
    peak_list = np.genfromtxt(args.path_list_peak, dtype=None, encoding="utf-8")

    # Read z values
    list_z = np.genfromtxt(args.path_list_z, dtype=None)

    # Create the output directory
    args.path_output.mkdir(parents=True, exist_ok=True)

    # Create unit conversion object for indirect and direct dimension
    ucy = ng.pipe.make_uc(dic, spectra, dim=1)
    ucx = ng.pipe.make_uc(dic, spectra, dim=2)

    len_z, len_y, len_x = spectra.shape

    # Define the active spectral regions for the fit
    mask = np.zeros_like(spectra)
    mask = clustering.mark_peaks(mask, peak_list, ucx, ucy)
    mask = clustering.find_independent_regions(spectra, mask, args.contour_level)

    # Cluster peaks
    clusters = clustering.cluster_peaks(spectra, mask, peak_list, ucx, ucy)

    #
    # Lineshape Fitting
    #

    print("- Lineshape fitting...", end="\n\n\n")

    string_amplitudes = []

    file_logs = (args.path_output / "logs.out").open("w")

    for peaks, grid_x, grid_y, data_to_fit in clusters:

        print_peaks(peaks, files=(sys.stdout, file_logs))

        n_peaks = len(peaks)
        params = shapes.create_params(peaks)
        args_fit = (grid_x, grid_y, data_to_fit, n_peaks, args.noise)

        out = lf.minimize(computing.residuals, params, args=args_fit)

        for name, param in out.params.items():
            if "eta" in name:
                param.set(min=param.min, max=param.max, vary=True)

        out = lf.minimize(computing.residuals, out.params, args=args_fit)

        message = lf.fit_report(out)
        print(message, end="\n\n\n")
        print(message, end="\n\n\n", file=file_logs)

        _shapes, amplitudes = computing.calculate_shape_amplitudes(
            out.params, *(args_fit[:4])
        )

        data_sim = computing.simulate_data(
            out.params, grid_x, grid_y, data_to_fit, n_peaks
        )

        #
        # Monte-Carlo
        #

        if args.noise_box and int(args.noise_box[4]) > 1:

            from random import sample

            noise_box = args.noise_box[:4]
            n_iter = int(args.noise_box[4])

            bounds_x = sorted(ucx.i(bound_x) for bound_x in noise_box[0:2])
            bounds_y = sorted(ucy.i(bound_y) for bound_y in noise_box[2:4])

            grid_for_noise = monte_carlo.get_noise_grid(
                bounds_x, bounds_y, grid_x, grid_y
            )

            params_mc_list = []
            amplitudes_mc_list = []

            for x, y in sample(grid_for_noise, n_iter):
                x_noise_list = grid_x - min(grid_x) + x
                y_noise_list = grid_y - min(grid_y) + y

                data_noise = (
                    spectra[:, y_noise_list, x_noise_list]
                    .reshape((len_z, grid_x.size))
                    .T
                )

                data_mc = data_sim + data_noise
                args_mc = (grid_x, grid_y, data_mc, n_peaks)
                out_mc = lf.minimize(computing.residuals, out.params, args=args_mc)

                _shapes_mc, amplitudes_mc = computing.calculate_shape_amplitudes(
                    out_mc.params, *args_mc
                )

                params_mc_list.append(out_mc.params.valuesdict())
                amplitudes_mc_list.append(amplitudes_mc)

            params_err = calc_err_from_mc(params_mc_list)
            amplitudes_err = np.std(amplitudes_mc_list, axis=0, ddof=1)

        else:

            params_err = {param.name: param.stderr for param in out.params.values()}
            amplitudes_err = np.zeros_like(amplitudes) + args.noise

        for index, peak in enumerate(peaks):
            name, _, _, x_alias, y_alias = peak

            string_amplitudes.append("{:>15s}".format(name))

            x_scale_ppm = abs(ucx.ppm(1.0) - ucx.ppm(0.0))
            y_scale_ppm = abs(ucx.ppm(1.0) - ucx.ppm(0.0))
            x_scale_hz = abs(ucx.hz(1.0) - ucx.hz(0.0))
            y_scale_hz = abs(ucy.hz(1.0) - ucy.hz(0.0))

            prefix = "p{}_".format(index)

            x0 = out.params["".join([prefix, "x0"])].value
            y0 = out.params["".join([prefix, "y0"])].value
            x0_err = params_err["".join([prefix, "x0"])]
            y0_err = params_err["".join([prefix, "y0"])]

            x_fwhm = out.params["".join([prefix, "x_fwhm"])].value
            y_fwhm = out.params["".join([prefix, "y_fwhm"])].value
            x_fwhm_err = params_err["".join([prefix, "x_fwhm"])]
            y_fwhm_err = params_err["".join([prefix, "y_fwhm"])]

            x_eta = out.params["".join([prefix, "x_eta"])].value
            y_eta = out.params["".join([prefix, "y_eta"])].value
            x_eta_err = params_err["".join([prefix, "x_eta"])]
            y_eta_err = params_err["".join([prefix, "y_eta"])]

            amplitude_err = np.mean(amplitudes_err[index])

            filename = args.path_output / "".join([name, ".out"])

            with filename.open("w") as f:

                f.write("# Name: {:>10s}\n".format(name))
                f.write(
                    "# y_ppm: {:10.5f} {:10.5f}\n".format(
                        ucy.ppm(y0 + y_alias * len_y), y0_err * y_scale_ppm
                    )
                )
                f.write(
                    "# yw_hz: {:10.5f} {:10.5f}\n".format(
                        y_fwhm * y_scale_hz, y_fwhm_err * y_scale_hz
                    )
                )
                f.write("# y_eta: {:10.5f} {:10.5f}\n".format(y_eta, y_eta_err))
                f.write(
                    "# x_ppm: {:10.5f} {:10.5f}\n".format(
                        ucx.ppm(x0 + x_alias * len_x), x0_err * x_scale_ppm
                    )
                )
                f.write(
                    "# xw_hz: {:10.5f} {:10.5f}\n".format(
                        x_fwhm * x_scale_hz, x_fwhm_err * x_scale_hz
                    )
                )
                f.write("# x_eta: {:10.5f} {:10.5f}\n".format(x_eta, x_eta_err))

                f.write("#---------------------------------------------\n")
                f.write("# {:>10s}  {:>14s}  {:>14s}\n".format("Z", "I", "I_err"))

                for z, amplitude in zip(list_z, amplitudes[index]):
                    f.write(
                        "  {:>10s}  {:14.6e}  {:14.6e}\n".format(
                            str(z), amplitude, amplitude_err
                        )
                    )
                    string_amplitudes.append("{:14.6e}".format(amplitude))

                string_amplitudes.append("\n")

    file_logs.close()

    filename = args.path_output / "amplitudes.out"
    filename.write_text(" ".join(string_amplitudes))


def calc_err_from_mc(params_mc_list):
    params_list = {}

    for params_mc in params_mc_list:

        for key, val in params_mc.items():
            params_list.setdefault(key, []).append(val)

    params_err = {key: np.std(val_list) for key, val_list in params_list.items()}

    return params_err


if __name__ == "__main__":
    main()



import collections
import itertools

import numpy as np

from peakfit import util

Peak = collections.namedtuple("Peak", ["name", "x_pt", "y_pt", "x_alias", "y_alias"])


def flood_fill(data, mask, threshold, cluster_id, x_start, y_start):
    ny, nx = data.shape

    stack = [(x_start, y_start)]

    while stack:

        x, y = stack.pop()

        if (abs(data[y, x]) >= threshold and mask[0, y, x] == 0.0) or mask[
            0, y, x
        ] < 0.0:

            mask[:, y, x] = cluster_id

            if x > 0:
                stack.append((x - 1, y))

            if x < (nx - 1):
                stack.append((x + 1, y))

            if y > 0:
                stack.append((x, y - 1))

            if y < (ny - 1):
                stack.append((x, y + 1))


def mark_peaks(mask, peak_list, ucx, ucy):
    """Marks peaks in mask"""

    print("- Marking peaks...")

    nz, ny, nx = mask.shape

    for peak in peak_list:

        x_index = ucx.i(peak[2], "ppm") % nx
        y_index = ucy.i(peak[1], "ppm") % ny

        x_index = util.clamp(x_index, 2, nx - 2)
        y_index = util.clamp(y_index, 2, ny - 2)

        mask[:, y_index - 2 : y_index + 3, x_index - 2 : x_index + 3] = -1.0

    return mask


def find_independent_regions(data, mask, contour_level):
    """Finds regions of the spectra above the contour level threshold."""

    print("- Segmenting the spectra according to the threshold level...")

    nz, ny, nx = data.shape

    cluster_index = 0

    for y, x in itertools.product(range(ny), range(nx)):

        if mask[0, y, x] <= 0.0:
            cluster_index += 1
            flood_fill(data[0], mask, contour_level, cluster_index, x, y)

    return mask


def cluster_peaks(spectra, mask, peak_list, ucx, ucy):
    """Identifies overlapping peaks."""

    print("- Clustering of peaks...")

    nz, ny, nx = mask.shape

    names = peak_list["f0"]
    x_pts = ucx.f(peak_list["f2"], "ppm")
    y_pts = ucy.f(peak_list["f1"], "ppm")
    x_aliases = x_pts // nx
    y_aliases = y_pts // ny
    x_pts %= nx
    y_pts %= ny
    peaks = zip(names, x_pts, y_pts, x_aliases, y_aliases)

    cluster_ids = list(mask[0, np.rint(y_pts).astype(int), np.rint(x_pts).astype(int)])

    peak_clusters = dict()
    for cluster_id, peak in zip(cluster_ids, peaks):
        peak_clusters.setdefault(cluster_id, []).append(peak)

    clusters = []

    for cluster_id, peak_cluster in peak_clusters.items():
        grid_y, grid_x = np.where(mask[0] == cluster_id)
        data_to_fit = spectra[:, grid_y, grid_x].reshape((nz, grid_x.size)).T
        clusters.append((peak_cluster, grid_x, grid_y, data_to_fit))

    return clusters



import itertools


def get_noise_grid(bounds_x, bounds_y, x_grid, y_grid):

    range_x = max(x_grid) - min(x_grid)
    range_y = max(y_grid) - min(y_grid)

    bounds_x[1] -= range_x
    bounds_y[1] -= range_y

    grid_noise = list(itertools.product(range(*bounds_x), range(*bounds_y)))

    return grid_noise



import argparse
import pathlib

import matplotlib.pyplot as plt
from natsort import natsorted
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages


def main():

    args = get_args()
    plot(args.files)

    return


def get_args():
    parser = argparse.ArgumentParser(description="Plot CEST profiles.")
    parser.add_argument("-f", "--files", nargs="+", type=pathlib.Path)
    return parser.parse_args()


def plot(files):

    figs = {}

    print()
    print("Reading files...")

    for a_file in files:

        print("  * {}".format(a_file.name))

        data = np.loadtxt(
            a_file,
            dtype={
                "names": ("offset", "intensity", "error"),
                "formats": ("i4", "f8", "f8"),
            },
        )

        data_ref = data[abs(data["offset"]) >= 1e4]
        data_cest = data[abs(data["offset"]) < 1e4]

        figs[a_file.name] = make_fig(a_file.name, data_cest, data_ref)

    print()
    print("Plotting...")

    with PdfPages("profiles.pdf") as pdf:
        for name in natsorted(figs):
            pdf.savefig(figs[name])

    print("  * profiles.pdf")


def offset_to_nu_cpmg(offset, time_t2):

    nu_cpmg = []

    for a_offset in offset:
        if a_offset > 0.0:
            nu_cpmg.append(a_offset / time_t2)
        else:
            nu_cpmg.append(0.5 / time_t2)

    return nu_cpmg


def intensity_to_r2eff(intensity, intensity_ref, time_t2):

    if np.any(intensity / intensity_ref <= 0.0):
        print(intensity / intensity_ref)
        print()

    return -np.log(intensity / intensity_ref) / time_t2


def make_ens(data, size=1000):

    return data["intensity"] + data["error"] * np.random.randn(
        size, len(data["intensity"])
    )


def make_fig(name, data_cest, data_ref):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.errorbar(
        data_cest["offset"],
        data_cest["intensity"] / np.mean(data_ref["intensity"]),
        yerr=data_cest["error"] / np.mean(data_ref["intensity"]),
        fmt="o",
    )
    ax.set_title(name)
    ax.set_xlabel(r"$B_1$ offset (Hz)")
    ax.set_ylabel(r"$I/I_0$")
    plt.close()

    return fig



import numpy as np
from scipy import linalg

from peakfit.shapes import pvoigt2d


def calculate_shape_amplitudes(
    params, x_list, y_list, data, n_peaks, amplitude_params=False
):
    shapes = []

    for index in range(n_peaks):
        prefix = "p{:d}_".format(index)
        x0 = params["".join([prefix, "x0"])].value
        y0 = params["".join([prefix, "y0"])].value
        x_fwhm = params["".join([prefix, "x_fwhm"])].value
        y_fwhm = params["".join([prefix, "y_fwhm"])].value
        x_eta = params["".join([prefix, "x_eta"])].value
        y_eta = params["".join([prefix, "y_eta"])].value

        shapes.append(pvoigt2d(x_list, y_list, x0, y0, x_fwhm, y_fwhm, x_eta, y_eta))

    shapes = np.asarray(shapes).T

    if amplitude_params:

        amplitudes = []

        for index in range(n_peaks):
            prefix = "p{:d}_".format(index)
            amplitudes.append(
                [params["".join([prefix, "i", str(nb)])] for nb in range(data.shape[1])]
            )

        amplitudes = np.asarray(amplitudes)

    else:

        amplitudes = linalg.lstsq(shapes, data)[0]

    return shapes, amplitudes


def residuals(params, x_list, y_list, data, n_peaks, noise=1.0, amplitude_params=False):
    shapes, amplitudes = calculate_shape_amplitudes(
        params, x_list, y_list, data, n_peaks, amplitude_params
    )

    return np.ravel((data - shapes.dot(amplitudes)) / noise)


def simulate_data(params, x_list, y_list, data, n_peaks):
    shapes, amplitudes = calculate_shape_amplitudes(
        params, x_list, y_list, data, n_peaks
    )

    return shapes.dot(amplitudes)



import argparse
import pathlib

import matplotlib.pyplot as plt
from natsort import natsorted
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages


def main():

    args = get_args()
    plot(args.files, args.time_t2)

    return


def get_args():
    parser = argparse.ArgumentParser(description="Plot CPMG R2eff profiles.")
    parser.add_argument("-f", "--files", nargs="+", type=pathlib.Path)
    parser.add_argument("-t", "--time_t2", type=float)
    return parser.parse_args()


def plot(files, time_t2):

    figs = {}

    print()
    print("Reading files...")

    for a_file in files:

        print("  * {}".format(a_file.name))

        data = np.loadtxt(
            a_file,
            dtype={
                "names": ("ncyc", "intensity", "error"),
                "formats": ("i4", "f8", "f8"),
            },
        )

        data_ref = data[data["ncyc"] == 0]
        data_cpmg = data[data["ncyc"] != 0]

        intensity_ref = np.mean(data_ref["intensity"])
        error_ref = np.mean(data_ref["error"]) / np.sqrt(len(data_ref))

        nu_cpmg = ncyc_to_nu_cpmg(data_cpmg["ncyc"], time_t2)

        r2_exp = intensity_to_r2eff(data_cpmg["intensity"], intensity_ref, time_t2)
        r2_ens = intensity_to_r2eff(
            make_ens(data_cpmg),
            make_ens(
                {"intensity": np.array([intensity_ref]), "error": np.array([error_ref])}
            ),
            time_t2,
        )

        r2_erd, r2_eru = abs(np.percentile(r2_ens, [15.9, 84.1], axis=0) - r2_exp)

        figs[a_file.name] = make_fig(a_file.name, nu_cpmg, r2_exp, r2_erd, r2_eru)

    print()
    print("Plotting...")

    with PdfPages("profiles.pdf") as pdf:
        for name in natsorted(figs):
            pdf.savefig(figs[name])

    print("  * profiles.pdf")


def ncyc_to_nu_cpmg(ncyc, time_t2):

    nu_cpmg = []

    for a_ncyc in ncyc:
        if a_ncyc > 0.0:
            nu_cpmg.append(a_ncyc / time_t2)
        else:
            nu_cpmg.append(0.5 / time_t2)

    return nu_cpmg


def intensity_to_r2eff(intensity, intensity_ref, time_t2):
    return -np.log(intensity / intensity_ref) / time_t2


def make_ens(data, size=1000):

    return data["intensity"] + data["error"] * np.random.randn(
        size, len(data["intensity"])
    )


def make_fig(name, nu_cpmg, r2_exp, r2_erd, r2_eru):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.errorbar(nu_cpmg, r2_exp, yerr=(r2_erd, r2_eru), fmt="o")
    ax.set_title(name)
    ax.set_xlabel(r"$\nu_{CPMG}$ (Hz)")
    ax.set_ylabel(r"$R_{2,eff}$ (s$^{-1}$)")
    plt.close()

    return fig



import lmfit as lf
import nmrglue.analysis.lineshapes1d as ls


def pvoigt2d(x, y, x0, y0, x_fwhm, y_fwhm, x_eta, y_eta):
    return ls.sim_pvoigt_fwhm(x, x0, x_fwhm, x_eta) * ls.sim_pvoigt_fwhm(
        y, y0, y_fwhm, y_eta
    )


def create_params(peaks):
    params = lf.Parameters()

    for index, peak in enumerate(peaks):
        prefix = "p{}_".format(index)
        x0, y0 = peak[1:3]

        x0_name = "".join([prefix, "x0"])
        y0_name = "".join([prefix, "y0"])
        x_fwhm_name = "".join([prefix, "x_fwhm"])
        y_fwhm_name = "".join([prefix, "y_fwhm"])
        x_eta_name = "".join([prefix, "x_eta"])
        y_eta_name = "".join([prefix, "y_eta"])

        params.add(x0_name, value=x0)
        params.add(y0_name, value=y0)
        params.add(x_fwhm_name, value=5.0, min=0.001, max=25.0)
        params.add(y_fwhm_name, value=5.0, min=0.001, max=25.0)
        params.add(x_eta_name, value=0.5, min=-1.0, max=+1.0, vary=False)
        params.add(y_eta_name, value=0.5, min=-1.0, max=+1.0, vary=False)

    return params



import errno
import os
import os.path
import sys


def make_dirs(path=None):
    """Make the directory if needed"""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def print_peaks(peaks, files=None):
    if files is None:
        files = (sys.stdout,)

    message = "*  Peak(s): "
    message += ", ".join(["{:s}".format(peak[0]) for peak in peaks])
    message += "  *"
    stars = "*" * len(message)
    message = "\n".join([stars, message, stars])

    for file in files:
        print(message, end="\n\n", file=file)


def clamp(n, smallest, largest):
    return max(smallest, min(n, largest))



from setuptools import find_packages, setup

import versioneer

with open("Readme.md") as f:
    long_description = f.read()

setup(
    name="PeakFit",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="PeakFit allow for lineshape fitting in pseudo-3D NMR spectra.",
    long_description=long_description,
    license="BSD 3-Clause",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    author="Guillaume Bouvignies",
    author_email="gbouvignies@gmail.com",
    packages=find_packages(),
    install_requires=[
        "numpy>=1.0",
        "scipy>=1.0",
        "matplotlib>=2.0",
        "natsort>=2.0",
        "nmrglue",
        "lmfit>=0.9.11",
        "asteval>=0.9.11",
    ],
    python_requires=">=3.5",
    entry_points={
        "console_scripts": [
            "peakfit = peakfit.peakfit:main",
            "plot_cpmg = peakfit.plot_cpmg:main",
            "plot_cest = peakfit.plot_cest:main",
        ]
    },
)
